{-# OPTIONS -cpp -fglasgow-exts #-}

module Context
(
    Context(Context#),
    ContextCombo(ContextCombo#),
    Cost,
    PF(PF),

    productionsFor,
    unitCombo,
    rootContext,
    priorCost,
    applyTagContext,
    applyTagContextCombo,
    spaceCombos,
    sameSpaceContext,
    sameSpaceContextCombo,
    isTT,
    pathTo,
    costTo
)
where

import PrelGHC
import PrelBase
import Addr
import Tag
import Cost

data Context      = Context#      !Word#
data ContextCombo = ContextCombo# !Word#
data PF           = PF !ContextCombo !Cost deriving (Eq, Show)
data PFComponent  = PFC# !Word# !Int#

instance Eq Context where
    Context# x# == Context# y# = x# `eqWord#` y#
    Context# x# /= Context# y# = x# `neWord#` y#

instance Eq ContextCombo where
    ContextCombo# x# == ContextCombo# y# = x# `eqWord#` y#
    ContextCombo# x# /= ContextCombo# y# = x# `neWord#` y#

instance Show Context where
    showsPrec _ (Context# x#) = showBits# x#

instance Show ContextCombo where
    showsPrec _ (ContextCombo# x#) = showBits# x#

PFC# _ k1# `lePFC` PFC# _ k2# = k1# <=# k2#
PFC# _ k1# `gePFC` PFC# _ k2# = k1# >=# k2#
PFC# _ k1# `ltPFC` PFC# _ k2# = k1# <#  k2#
PFC# _ k1# `gtPFC` PFC# _ k2# = k1# >#  k2#

-- Context/ContextCombo bit pattern:
--                             _
--      31      TT              \            _
--      30      B                |- "enable"  \
--      29      I              _/              |_ "minimize"
--      27-28   volume base,base-or-em,s,em    |
--      25-26   U                            _/
--      16-24   colors rgbcmykw?  -.
--      11-15   unused             |- "update"
--      0-10    sizes 0-9,?       -'

#define MASK_TT         ( int2Word# 0x80000000# )
#define MASK_B          ( int2Word# 0x40000000# )
#define MASK_I          ( int2Word# 0x20000000# )
#define MASK_VolumeBase ( int2Word# 0x00000000# )
#define MASK_VolumeWeak ( int2Word# 0x08000000# )
#define MASK_VolumeS    ( int2Word# 0x10000000# )
#define MASK_VolumeEM   ( int2Word# 0x18000000# )
#define MASK_VolumeAny  ( int2Word# 0x18000000# )
#define BITS_Volume     27
#define MASK_U3         ( int2Word# 0x06000000# )
#define MASK_U2         ( int2Word# 0x04000000# )
#define MASK_U1         ( int2Word# 0x02000000# )
#define MASK_U0         ( int2Word# 0x00000000# )
#define MASK_UAny       ( MASK_U2 `or#` MASK_U1 )
#define BITS_U          25
#define MASK_ColorRoot  ( int2Word# 0x01000000# )
#define MASK_ColorSpec  ( int2Word# 0x00ff0000# )
#define MASK_ColorR     ( int2Word# 0x00010000# )
#define MASK_ColorG     ( int2Word# 0x00020000# )
#define MASK_ColorB     ( int2Word# 0x00040000# )
#define MASK_ColorC     ( int2Word# 0x00080000# )
#define MASK_ColorM     ( int2Word# 0x00100000# )
#define MASK_ColorY     ( int2Word# 0x00200000# )
#define MASK_ColorK     ( int2Word# 0x00400000# )
#define MASK_ColorW     ( int2Word# 0x00800000# )
#define MASK_ColorAny   ( MASK_ColorRoot `or#` MASK_ColorSpec )
#define MASK_SizeRoot   ( int2Word# 0x00000400# )
#define MASK_SizeSpec   ( int2Word# 0x000003ff# )
#define MASK_Size0      ( int2Word# 0x00000001# )
#define MASK_Size1      ( int2Word# 0x00000002# )
#define MASK_Size2      ( int2Word# 0x00000004# )
#define MASK_Size3      ( int2Word# 0x00000008# )
#define MASK_Size4      ( int2Word# 0x00000010# )
#define MASK_Size5      ( int2Word# 0x00000020# )
#define MASK_Size6      ( int2Word# 0x00000040# )
#define MASK_Size7      ( int2Word# 0x00000080# )
#define MASK_Size8      ( int2Word# 0x00000100# )
#define MASK_Size9      ( int2Word# 0x00000200# )
#define MASK_SizeAny    ( MASK_SizeRoot `or#` MASK_SizeSpec )
#define MASK_Zero       ( int2Word# 0# )
#define MASK_Enable     ( MASK_TT `or#` MASK_B `or#` MASK_I )
#define MASK_Minimize   ( MASK_Enable `or#` MASK_VolumeAny `or#` MASK_UAny )
#define MASK_Update     ( MASK_ColorAny `or#` MASK_SizeAny )

maskU#, extractU#, maskVolume#, extractVolume# :: Word# -> Word#
maskU#         x# = x# `and#` MASK_UAny
extractU#      x# = maskU# x# `shiftRL#` BITS_U#
maskVolume#    x# = x# `and#` MASK_VolumeAny
extractVolume# x# = maskVolume# x# `shiftRL#` BITS_Volume#

showBits# :: Word# -> ShowS
showBits# x# = showChar '{' . foldr1 punctuate attributes . showChar '}'
  where
    punctuate f g = f . showChar ',' . g
    attributes = [ showString "TT" | test# x# MASK_TT ]
              ++ [ showString "B"  | test# x# MASK_B  ]
              ++ [ showString "I"  | test# x# MASK_I  ]
              ++ [ showString "EM?" | maskVolume# x# `eqWord#` MASK_VolumeWeak ]
              ++ [ showString "S"   | maskVolume# x# `eqWord#` MASK_VolumeS    ]
              ++ [ showString "EM"  | maskVolume# x# `eqWord#` MASK_VolumeEM   ]
              ++ [ (if test# x# MASK_U2 then showString "UU" else id)
                 . (if test# x# MASK_U1 then showString "U"  else id)
                 | test# x# MASK_UAny ]
              ++ [ foldr1 (.) $ map showChar $
                   [ 'r' | test# x# MASK_ColorR ] ++
                   [ 'g' | test# x# MASK_ColorG ] ++
                   [ 'b' | test# x# MASK_ColorB ] ++
                   [ 'c' | test# x# MASK_ColorC ] ++
                   [ 'm' | test# x# MASK_ColorM ] ++
                   [ 'y' | test# x# MASK_ColorY ] ++
                   [ 'k' | test# x# MASK_ColorK ] ++
                   [ 'w' | test# x# MASK_ColorW ] ++
                   [ '^' | test# x# MASK_ColorRoot ]
                 , foldr1 (.) $ map showChar $
                   [ '0' | x# `test#` MASK_Size0 ] ++
                   [ '1' | x# `test#` MASK_Size1 ] ++
                   [ '2' | x# `test#` MASK_Size2 ] ++
                   [ '3' | x# `test#` MASK_Size3 ] ++
                   [ '4' | x# `test#` MASK_Size4 ] ++
                   [ '5' | x# `test#` MASK_Size5 ] ++
                   [ '6' | x# `test#` MASK_Size6 ] ++
                   [ '7' | x# `test#` MASK_Size7 ] ++
                   [ '8' | x# `test#` MASK_Size8 ] ++
                   [ '9' | x# `test#` MASK_Size9 ] ++
                   [ '^' | x# `test#` MASK_SizeRoot ] ]

#define TABLE(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p) \
    ( (int2Word# a# `shiftL#`  0#) `or#` \
      (int2Word# b# `shiftL#`  2#) `or#` \
      (int2Word# c# `shiftL#`  4#) `or#` \
      (int2Word# d# `shiftL#`  6#) `or#` \
      (int2Word# e# `shiftL#`  8#) `or#` \
      (int2Word# f# `shiftL#` 10#) `or#` \
      (int2Word# g# `shiftL#` 12#) `or#` \
      (int2Word# h# `shiftL#` 14#) `or#` \
      (int2Word# i# `shiftL#` 16#) `or#` \
      (int2Word# j# `shiftL#` 18#) `or#` \
      (int2Word# k# `shiftL#` 20#) `or#` \
      (int2Word# l# `shiftL#` 22#) `or#` \
      (int2Word# m# `shiftL#` 24#) `or#` \
      (int2Word# n# `shiftL#` 26#) `or#` \
      (int2Word# o# `shiftL#` 28#) `or#` \
      (int2Word# p# `shiftL#` 30#) )

productionsFor :: ContextCombo -> ContextCombo -> [PF]
productionsFor (ContextCombo# l#) (ContextCombo# r#)
    = [ PF (ContextCombo# x#) (Cost# k#)
      | PFC# x# k# <- prodSize `mergePFC` prodColor `mergePFC` prodMinimize ]

  where
    prodSize  = prodUpdate MASK_SizeAny MASK_SizeRoot
    prodColor = prodUpdate MASK_ColorAny MASK_ColorRoot

    prodUpdate mask# root# =
      case l# `and#` mask# of { lField# ->
      case r# `and#` mask# of { rField# ->
      if lField# `eqWord#` mask# then [ PFC# rField# 0# ] else
      if rField# `eqWord#` mask# then [ PFC# lField# 0# ] else
      if lField# `eqWord#` rField# then [ PFC# lField# 0# ] else
      case test# root# (l# `or#` r#) of { hasRoot -> filter nonEmpty
      [ PFC# (lField# `and#` rField#) 0#
      , PFC# (if hasRoot then root# else lField# `xor#` rField#) 7# ] } } }
    nonEmpty (PFC# x# _) = x# `neWord#` MASK_Zero

    prodMinimize =
      case minAndCost# l# r# of { (# min#, cost# #) ->
      case l# `and#` MASK_Minimize of { lMinimize# ->
      case r# `and#` MASK_Minimize of { rMinimize# ->
      if min# `eqWord#` lMinimize#
      then if min# `eqWord#` rMinimize#
           then [ PFC# min# cost# ]
           else sortPFC2 ( PFC# min# cost# )
                         ( PFC# rMinimize# (9# +# priorCost_Minimize# l#) )
      else if min# `eqWord#` rMinimize#
           then sortPFC2 ( PFC# min# cost# )
                         ( PFC# lMinimize# (9# +# priorCost_Minimize# r#) )
           else sortPFC3 ( PFC# min# cost# )
                         ( PFC# lMinimize# (9# +# priorCost_Minimize# r#) )
                         ( PFC# rMinimize# (9# +# priorCost_Minimize# l#) ) }}}

addPFC :: PFComponent -> PFComponent -> PFComponent
addPFC (PFC# x1# k1#) (PFC# x2# k2#) = PFC# (x1# `or#` x2#) (k1# +# k2#)

mergePFC :: [PFComponent] -> [PFComponent] -> [PFComponent]
mergePFC comps1 comps2
  = foldr1 merge [ map (addPFC comp1) comps2 | comp1 <- comps1 ]
  where
    merge [] ys       = ys
    merge xs []       = xs
    merge xs@(x:xt) ys@(y:yt)
        | x `lePFC` y = x : merge xt ys
        | otherwise   = y : merge xs yt

sortPFC2 x y | x `lePFC` y = [x, y]
             | otherwise   = [y, x]

sortPFC3 x y z = if x `lePFC` y 
                 then if y `lePFC` z
                      then [x, y, z]
                      else if x `lePFC` z
                           then [x, z, y]
                           else [z, x, y]
                 else if x `lePFC` z
                      then [y, x, z]
                      else if y `lePFC` z
                           then [y, z, x]
                           else [z, y, x]

minAndCost# :: Word# -> Word# -> (# Word#, Int# #)
minAndCost# l# r#
  = (# enableMin# `or#` volumeMin# `or#` underlineMin#
     , enableCost#  +#  volumeCost#  +#  underlineCost# #)

  where
    volumeMinTable#      = TABLE(0,0,0,1, 0,1,1,3, 0,1,2,3, 1,3,3,3)
    underlineMinTable#   = TABLE(0,0,0,0, 0,1,1,1, 0,1,2,2, 0,1,2,3)
    volumeCostTable#     = TABLE(0,0,1,2, 0,0,1,0, 1,1,0,1, 2,0,1,0)
    underlineCostTable#  = TABLE(0,1,2,3, 1,0,1,2, 2,1,0,1, 3,2,1,0)

    volumeIndex#    = index# BITS_Volume#
    underlineIndex# = index# BITS_U#
    index# :: Int# -> Int#
    index# bits# = word2Int#
        ( or# ((l# `shiftRL#` (bits# -# 1#)) `and#` (int2Word# 0x06#))
              ((r# `shiftRL#` (bits# -# 3#)) `and#` (int2Word# 0x18#)) )

    volumeMin#     = lookup# volumeMinTable#    volumeIndex#
                     `shiftL#` BITS_Volume#
    underlineMin#  = lookup# underlineMinTable# underlineIndex#
                     `shiftL#` BITS_U#
    volumeCost#    = case word2Int# (lookup# volumeCostTable# volumeIndex#) of
                      { 0# -> 0#; 1# -> 7#; 2# -> 9# }
    underlineCost# = word2Int# (lookup# underlineCostTable# underlineIndex#)
                     *# 7#
    lookup# :: Word# -> Int# -> Word#
    lookup# table# idx# = (table# `shiftRL#` idx#) `and#` (int2Word# 3#)

    enableMin#  = l# `and#` r# `and#` MASK_Enable
    enableCost# = priorCost_Enable# (l# `xor#` r#)

unitCombo :: Context -> ContextCombo
unitCombo (Context# x) = ContextCombo# x

rootContext :: Context
rootContext = Context#
    (MASK_VolumeBase `or#` MASK_ColorRoot `or#` MASK_SizeRoot)

priorCost :: ContextCombo -> Cost
priorCost (ContextCombo# x#) = Cost#
     ( priorCost_Minimize# x#
    +# (if test# MASK_ColorRoot x# then 0# else 7#)
    +# (if test# MASK_SizeRoot  x# then 0# else 7#) )

priorCost_Minimize# :: Word# -> Int#
priorCost_Minimize# x#
    =  priorCost_Enable# x#
    +# priorCost_Volume# x#
    +# priorCost_Underline# x#

priorCost_Enable# :: Word# -> Int#
priorCost_Enable# x#
    =  testWeight# MASK_TT 9# x#
    +# testWeight# MASK_B  7# x#
    +# testWeight# MASK_I  7# x#

priorCost_Volume# :: Word# -> Int#
priorCost_Volume# x#
    = word2Int#
    ( and# (int2Word# 0xff#)
           ( shiftRL# (int2Word# 0x09070000#)
                      ( word2Int# ( shiftRL# (maskVolume# x#)
                                             (BITS_Volume# -# 3#) ) ) ) )

priorCost_Underline# :: Word# -> Int#
priorCost_Underline# x# = word2Int# (extractU# x#) *# 7#

applyTagContext :: Tag -> Context -> Context
applyTagContext tag (Context# x#) = Context# (applyTag# tag x#)

applyTagContextCombo :: Tag -> ContextCombo -> ContextCombo
applyTagContextCombo tag (ContextCombo# x#) = ContextCombo# (applyTag# tag x#)

applyTag# :: Tag -> Word# -> Word#
applyTag# TB     x# = x# `or#` MASK_B
applyTag# TI     x# = x# `or#` MASK_I
applyTag# TTT    x# = x# `or#` MASK_TT
applyTag# TPL    x# = x# `and#` not# MASK_Minimize
applyTag# TS     x# = rest# `or#` MASK_VolumeS
    where rest#   = x# `and#` not# MASK_VolumeAny
applyTag# TEM    x#
    | volume# `eqWord#` MASK_VolumeEM   = rest# `or#` MASK_VolumeBase
    | volume# `eqWord#` MASK_VolumeBase = rest# `or#` MASK_VolumeEM
    | otherwise                         = x#
    where volume# = x# `and#`      MASK_VolumeAny
          rest#   = x# `and#` not# MASK_VolumeAny
applyTag# TU     x#
    | maskU# x# `eqWord#` MASK_U3  = x#
    | otherwise  = int2Word# (word2Int# x# +# word2Int# MASK_U1)
applyTag# (S size#) x#
    | size# >=# 0# && size# <=# 9# -- TODO remove? robust?
    = (rest# `or#` (MASK_Size0 `shiftL#` size#))
    where rest#   = x# `and#` not# MASK_SizeAny
applyTag# (C color) x#
    = or# rest#
        ( case color of Cr -> MASK_ColorR
                        Cg -> MASK_ColorG
                        Cb -> MASK_ColorB
                        Cc -> MASK_ColorC
                        Cm -> MASK_ColorM
                        Cy -> MASK_ColorY
                        Ck -> MASK_ColorK
                        Cw -> MASK_ColorW )
    where rest#   = x# `and#` not# MASK_ColorAny

spaceCombos :: ContextCombo -> [ContextCombo]
spaceCombos (ContextCombo# x#)
    = [ ContextCombo# (y1#)
      , ContextCombo# (y2#)
      , ContextCombo# (y1# `or#` (MASK_B))
      , ContextCombo# (y1# `or#` (MASK_I))
      , ContextCombo# (y2# `or#` (MASK_B))
      , ContextCombo# (y2# `or#` (MASK_I))
      , ContextCombo# (y1# `or#` (MASK_B `or#` MASK_I))
      , ContextCombo# (y2# `or#` (MASK_B `or#` MASK_I)) ]
    where y#  = spaceContext# x#
          y1# = y# `or#` MASK_VolumeWeak
          y2# = y# `or#` MASK_VolumeS

sameSpaceContext :: Context -> Context -> Bool
sameSpaceContext (Context# x1#) (Context# x2#)
    = spaceContext# x1# `eqWord#` spaceContext# x2#

sameSpaceContextCombo :: ContextCombo -> ContextCombo -> Bool
sameSpaceContextCombo (ContextCombo# x1#) (ContextCombo# x2#)
    = spaceContext# x1# `eqWord#` spaceContext# x2#

spaceContext# :: Word# -> Word#
spaceContext# x#
    = and# (if (x# `test#` MASK_UAny) then x# else (x# `or#` MASK_ColorAny))
           (MASK_UAny `or#` MASK_TT `or#` MASK_ColorAny `or#` MASK_SizeAny)

isTT :: ContextCombo -> Bool
isTT (ContextCombo# x#) = test# MASK_TT x#

costTo :: Context -> ContextCombo -> Cost
from `costTo` to
    = Cost# (costOfTags (fst (from `pathTo` to)))
    where costOfTags []         = 0#
          costOfTags (tag:tags) = costOfTag tag +# costOfTags tags
          costOfTag  TEM        = 9#
          costOfTag  TPL        = 9#
          costOfTag  TTT        = 9#
          costOfTag  _          = 7#

pathTo :: Context -> ContextCombo -> ([Tag], Context)
Context# from# `pathTo` ContextCombo# to#
    = case pathTo_Minimize# from# to# of { (# tagsMinimize, resultMinimize #) ->
      case pathTo_Color#    from# to# of { (# tagsColor   , resultColor    #) ->
      case pathTo_Size#     from# to# of { (# tagsSize    , resultSize     #) ->
      ( tagsMinimize ++ tagsColor ++ tagsSize
      , Context# (resultMinimize `or#` resultColor `or#` resultSize) ) } } }

pathTo_Minimize# :: Word# -> Word# -> (# [Tag], Word# #)
from# `pathTo_Minimize#` to#
    | needPL     = case startAt MASK_Zero of
                   (# tags, result #) -> (# TPL:tags, result #)
    | otherwise  = startAt from#

  where
    fromEnable# = from# `and#` MASK_Enable
    fromU# = word2Int# (maskU# from#)
    toU#   = word2Int# (maskU# to#  )
    needPL = (fromEnable# `and#` to#) `neWord#` fromEnable#
              || fromU# ># toU#
              || ( maskVolume# from# `eqWord#` MASK_VolumeS
                && maskVolume# to#   `neWord#` MASK_VolumeS )

    startAt :: Word# -> (# [Tag], Word# #)
    startAt st#
        = (# [ TTT | diffTT ] ++
             [ TB  | diffB  ] ++
             [ TI  | diffI  ] ++
             tagsU ++ tagsVolume
           , toEnable# `or#` int2Word# toU# `or#` resultVolume# #)
      where
        toEnable#   = to# `and#` MASK_Enable
        stEnable#   = st# `and#` MASK_Enable
        diffEnable# = stEnable# `xor#` toEnable#                :: Word#
        diffTT      = diffEnable# `test#` MASK_TT               :: Bool
        diffB       = diffEnable# `test#` MASK_B                :: Bool
        diffI       = diffEnable# `test#` MASK_I                :: Bool
        stU#        = word2Int# (maskU# st#)                    :: Int#
        diffU#      = (toU# -# stU#) `iShiftRL#` BITS_U#        :: Int#
        tagsU       = replicate (I# diffU#) TU                  :: [Tag]
        stVolume#   = maskVolume# st#                           :: Word#
        toVolume#   = maskVolume# to#                           :: Word#
        ( tagsVolume, Context# resultVolume# )          :: ([Tag], Context)
            | toVolume# `eqWord#` MASK_VolumeS
            = ( if stVolume# `eqWord#` MASK_VolumeS then [] else [TS]
              , Context# MASK_VolumeS )
            | stVolume# `eqWord#` MASK_VolumeBase
            = if toVolume# `eqWord#` MASK_VolumeEM
              then ([TEM], Context# MASK_VolumeEM)
              else ([   ], Context# stVolume#  )
            | stVolume# `eqWord#` MASK_VolumeEM
            = if toVolume# `eqWord#` MASK_VolumeBase
              then ([TEM], Context# MASK_VolumeBase)
              else ([   ], Context# stVolume#    )

pathTo_Color# :: Word# -> Word# -> (# [Tag], Word# #)
from# `pathTo_Color#` to#
    | test# fromColor#  to#  = (# [    ], fromColor#  #)
    | test# MASK_ColorR to#  = (# [C Cr], MASK_ColorR #)
    | test# MASK_ColorG to#  = (# [C Cg], MASK_ColorG #)
    | test# MASK_ColorB to#  = (# [C Cb], MASK_ColorB #)
    | test# MASK_ColorC to#  = (# [C Cc], MASK_ColorC #)
    | test# MASK_ColorM to#  = (# [C Cm], MASK_ColorM #)
    | test# MASK_ColorY to#  = (# [C Cy], MASK_ColorY #)
    | test# MASK_ColorK to#  = (# [C Ck], MASK_ColorK #)
    | test# MASK_ColorW to#  = (# [C Cw], MASK_ColorW #)
    where fromColor# = from# `and#` MASK_ColorAny

pathTo_Size# :: Word# -> Word# -> (# [Tag], Word# #)
from# `pathTo_Size#` to#
    | test# fromSize#  to#  = (# [    ], fromSize#  #)
    | test# MASK_Size0 to#  = (# [S 0#], MASK_Size0 #)
    | test# MASK_Size1 to#  = (# [S 1#], MASK_Size1 #)
    | test# MASK_Size2 to#  = (# [S 2#], MASK_Size2 #)
    | test# MASK_Size3 to#  = (# [S 3#], MASK_Size3 #)
    | test# MASK_Size4 to#  = (# [S 4#], MASK_Size4 #)
    | test# MASK_Size5 to#  = (# [S 5#], MASK_Size5 #)
    | test# MASK_Size6 to#  = (# [S 6#], MASK_Size6 #)
    | test# MASK_Size7 to#  = (# [S 7#], MASK_Size7 #)
    | test# MASK_Size8 to#  = (# [S 8#], MASK_Size8 #)
    | test# MASK_Size9 to#  = (# [S 9#], MASK_Size9 #)
    where fromSize# = from# `and#` MASK_SizeAny

test# :: Word# -> Word# -> Bool
test# x# y# = (x# `and#` y#) `neWord#` MASK_Zero

testWeight# :: Word# -> Int# -> Word# -> Int#
testWeight# mask# weight# x# | x# `test#` mask# = weight#
                             | otherwise        = 0#
