module Driver where

import Concurrent
import MVar
import Posix (rename, getFileStatus, fileSize)
import System (getEnv)
import Monad (liftM, when)
import IO

driver :: ((String -> IO ()) -> IO ()) -> IO ()
driver computeInto = do

  debugging <- catch (getEnv "DEBUGGING" >> return True) (\e -> return False)

  tempFilePath <- catch (getEnv "TEMP_FILE_PATH") (\e -> return "/tmp/temp")
  bestFilePath <- catch (getEnv "BEST_FILE_PATH") (\e -> return "/tmp/best")
  initialBadness <- catch (liftM fileSize (getFileStatus bestFilePath))
                          (\e -> return 0)

  sv <- newEmptyMVar
  forkIO $ do
    catch (computeInto (\soln -> do putMVar sv (Just soln)))
          (\e -> hPutStrLn stderr ("Worker thread exception: " ++ show e))
    putMVar sv Nothing
  let samplingLoop :: Int -> IO ()
      samplingLoop minBadness = do
        maybeNewBadness <-
          catch (do when debugging $
                      hPutStrLn stderr ("minBadness = " ++ show minBadness)
                    maybeValue <- takeMVar sv
                    case maybeValue of
                      Nothing -> do
                        when debugging $
                          hPutStrLn stderr "End of solution stream"
                        return Nothing
                      Just value -> do
                        when debugging $
                          hPutStrLn stderr "Received solution"
                        hTemp <- openFile tempFilePath WriteMode
                        hPutStr hTemp value
                        curBadness <- liftM fromIntegral (hFileSize hTemp)
                        when debugging $
                          hPutStrLn stderr ("curBadness = " ++ show curBadness)
                        hClose hTemp
                        if curBadness < minBadness
                          then do when debugging $ hPutStrLn stderr "Yay!"
                                  rename tempFilePath bestFilePath
                                  return (Just curBadness)
                          else do return (Just minBadness))
                (\e -> do hPutStrLn stderr
                            ("Sampling thread exception: " ++ show e)
                          threadDelay 5000000
                          return (Just minBadness))
        case maybeNewBadness of
          Nothing -> return ()
          Just newBadness -> do
            samplingLoop newBadness
  samplingLoop (if initialBadness > 0 then initialBadness else maxBound)

computeInto reportSolution = computeInto' 10 where
    computeInto' 0 = return ()
    computeInto' n = do reportSolution (replicate n '*')
                        threadDelay 1000000
                        computeInto' (n-1)

