module Main where
import Parse
import Lex
import Random
import Tag

import IO
import qualified System

data Gen1Token = GOpen Tag | GClose | GLit | GSpace
  deriving (Eq, Show)

gen1weights :: [(Int, Gen1Token)]
gen1weights = zip (repeat 2) (map GOpen allTags) ++
		   [(length allTags * 2 + 1, GClose),
		    (40, GLit),
		    (40, GSpace)]
gen1total = sum (map fst gen1weights) - 1

gen1tok :: Int -> Gen1Token
gen1tok i = gen1tok' i gen1weights
gen1tok' i ((j,t):ts) | i < j     = t
		      | otherwise = gen1tok' (i-j) ts

gen1seq :: StdGen -> ([Gen1Token], StdGen)
gen1seq rg = let (rg', rg'') = split rg
		 rs = randomRs (0,gen1total) rg''
             in (map gen1tok rs, rg')

gen1lex :: Int -> [Gen1Token] -> [Lexeme]
gen1lex i toks = gen1lex' i [] toks

gen1lex' i [] (GClose:toks) | i <= 0    = []
			    | otherwise = gen1lex' i [] toks
gen1lex' i (t:ts) (GClose:toks) = Close t:gen1lex' (i-1) ts toks
gen1lex' i ts (GOpen t:toks) = Open t:gen1lex' (i-1) (t:ts) toks
gen1lex' i ts (GLit:toks) = Lit ((show (i`mod`8)),False):gen1lex' (i-1) ts toks
gen1lex' i ts (GSpace:toks) = Lit (" ",True):gen1lex' (i-1) ts toks

runGen1 :: Int -> IO [Lexeme]
runGen1 i = do toks <- getStdRandom gen1seq
	       return (gen1lex i toks)

main = do
   [length] <- System.getArgs
   let l = (read length) :: Int
   dat <- runGen1 l
   putStr (unLex dat)
