module Lex where
import Prelude hiding (lex)
import Tag

-- the Boolean is "true" if the literal is purely white space
type Literal = (String, Bool)

data Lexeme = Lit Literal | Open Tag | Close Tag
  deriving (Eq, Show)

isSpace :: Char -> Bool
isSpace ' ' = True
isSpace '\r' = True
isSpace '\n' = True
isSpace '\t' = True
isSpace _ = False

isNormal :: Char -> Bool
isNormal '<' = False
isNormal c | isSpace c = False
isNormal _ = True

lex :: [Char] -> [Lexeme]
lex [] = []
lex ('<':s) =
   case s of
     '/':s' -> case reads s' of
	         [(tag,'>':s'')] -> Close tag:lex s''
		 _ -> error ("Couldn't lex close tag at " ++ take 20 s')
     s' -> case reads s' of
             [(tag,'>':s'')] -> Open tag:lex s''
	     _ -> error ("Couldn't lex open tag at " ++ take 20 s')
lex s@(w:_) | isSpace w = let (s', s'') = span isSpace s in
                          Lit (s',True):lex s''
lex s = let (s', s'') = span isNormal s in Lit (s', False):lex s''

mergeSpaces :: [Char] -> [Char]
mergeSpaces [] = []
mergeSpaces (c:cs) | isSpace c = ' ':mergeSpaces (dropWhile isSpace cs)
	           | otherwise = c:mergeSpaces cs

unLex :: [Lexeme] -> String
unLex [] = []
unLex (Lit (s,_):ls) = s++unLex ls
unLex (Open t:ls) = '<':shows t ('>':unLex ls)
unLex (Close t:ls) = '<':'/':shows t ('>':unLex ls)
