{-# OPTIONS -fglasgow-exts #-}

module Main(doPass, main) where

import Prelude hiding (lex)
import Util
import Viterbi
import DynProg
import Cell
import Lex
import Parse
import Cost
import qualified System
import PrelBase(Int(I#))
import Driver

doPass :: Int -> Threshold -> [Cell] -> [Cell]
doPass k thresh l =
   map (chopCell thresh)
      (viterbi
         (map (map (costCell thresh))
            (take k
               (dynProg (accumulateCell thresh) emptyCell
                  l))))

cellsToString :: Threshold -> [Cell] -> String
cellsToString thresh s = concat (map (unLex .
				      unParse .
				      value .
				      (bestParse thresh)) s)

main = driver main3

main1 reportSolution = do
   [ks, maxdiff, maxlen] <- System.getArgs
   let k0:theKs = repeatLast (read ks)
       I# maxDiff# = read maxdiff
       maxLen = read maxlen
       theThresh = Threshold# (Cost# maxDiff#) maxLen
   input <- getContents
   let ptree = parse (lex input)
   reportSolution (unLex (unParse ptree))
   let start :: [Cell]
       start = map litToCell (flatten ptree)
   let approxs = start:zipWith (\approx theK -> doPass theK theThresh approx)
		               approxs
			       theKs
       approxs' = chopTail approxs
   reportSolution (cellsToString theThresh (last approxs'))

main2 reportSolution = do
   [ks, maxdiff, maxlen] <- System.getArgs
   let k0:theKs = repeatLast (read ks)
       I# maxDiff# = read maxdiff
       maxLen = read maxlen
       theThresh = Threshold# (Cost# maxDiff#) maxLen
   input <- getContents
   let ptree = parse (lex input)
       start = map litToCell (flatten ptree)
   reportSolution (unLex (unParse ptree))
   let dyn0 = dynProg (accumulateCell theThresh) emptyCell start
       pass0s = [map (chopCell theThresh)
		    (viterbi
		       (map (map (costCell theThresh))
		          (take k dyn0))) | k <- [k0..]]
   sequence_ [do sol <- (computeApprox theKs theThresh pass0)
	         reportSolution sol
	      | pass0 <- pass0s]

main3 reportSolution = do
   [ks, maxdiff, maxlen] <- System.getArgs
   let k0:theKs = repeatLast (read ks)
       I# maxDiff# = read maxdiff
       maxLen = read maxlen
       theThresh = Threshold# (Cost# maxDiff#) maxLen
   input <- getContents
   let ptree = parse (lex input)
       start = map litToCell (flatten ptree)
   reportSolution (unLex (unParse ptree))
   sequence_ [do sol <- computeApprox (k:theKs) theThresh start
		 reportSolution sol
	      | k <- [k0,k0+2..]]

computeApprox :: [Int] -> Threshold -> [Cell] -> IO String
computeApprox ks thresh start =
    let approxs = start:zipWith (\approx k -> doPass k thresh approx)
		                approxs
			 	ks
	approxs' = chopTail approxs
    in do sequence_ (map sampleOut approxs')
	  return (cellsToString thresh (last approxs'))

chopTail :: [[a]] -> [[a]]
chopTail (xs:xss) = xs:chopTail' (length xs) xss
chopTail' l (xs:xss) | length xs == l = []
		     | otherwise      = xs:chopTail' (length xs) xss

theThresh = Threshold# (Cost# 10#) 10

sampleOut l = do putStr (show (length l) ++ " ")
		 putStr (show (sumCost (map (rootCost.(bestParse theThresh)) l)) ++ " ")
--		 putStrLn (show l)

sumCost :: [Cost] -> Cost
sumCost = foldl (+$) zeroCost

repeatLast :: [a] -> [a]
repeatLast [x] = repeat x
repeatLast (x:xs) = x:repeatLast xs
