{-# OPTIONS -fglasgow-exts #-}

module Main(
  World(..),
  buildMatrix,
  buildMatrix',
  main
) where

import Context
import Prelude hiding (lex)
import Lex
import Parse
import TriMatrix
import Cell
import Util
import Cost

type Terminals = [(ContextCombo, String)]
data World = World (TriMatrix Cell) Terminals

type BuildState = (TriMatrix Cell, Int)

buildMatrix :: ParseTree -> World
buildMatrix tree =
   let strs = map (\(cntxt, (lit, _)) -> (cntxt,lit)) (flatten tree)
       size = length strs
   in World (fst (buildMatrix' tree (emptyTM size (Cell []), 0))) strs

buildMatrix' :: ParseTree -> BuildState -> BuildState
buildMatrix' (PLit (cntxt, (_, False))) (mat, n) =
  (insertTM mat n (n+1) (Cell [CellData cntxt
					(-1) -- Made up values!!!
					cntxt
					cntxt
					zeroCost]),
   n+1)
buildMatrix' (PLit (cntxt, (_, True))) (mat, n) =
  (insertTM mat n (n+1) (Cell
    (map ( \cntxt' -> CellData cntxt'
			       (-1) -- Made up values!!!
			       cntxt'
			       cntxt'
			       zeroCost) (spaceCombos cntxt))),
   n+1)
buildMatrix' (Branch cntxt [t1,t2]) s@(_,i) =
  let s'@(_,j) = buildMatrix' t1 s
      s''@(mat,k) = buildMatrix' t2 s'
      cc1 = treeContextCombo t1
      cc2 = treeContextCombo t2
      cntxt' = snd (pathTo rootContext cntxt) -- UNSAFE!
  in (insertTM mat i k (Cell [CellData cntxt
				     j
				     cc1
				     cc2
				     (costTo cntxt' cc1 +$ costTo cntxt' cc2)]),
      k)
buildMatrix' t _ = error ("Invalid parse tree: " ++ show t)

bestParse :: World -> ParseTree
bestParse (World mat terminals) =
  let cell0 = lookupTM mat 0 (sizeTM mat)
      bestParse' :: ContextCombo -> Int -> Int -> Terminals
                           -> (ParseTree, Terminals)
      bestParse' c1 i j terms
       | j == i+1 = let ((c2, lit):terms') = terms in
                      (PLit (c2, (lit, False)), terms')
       | otherwise = let cd = findContextCombo c1 (lookupTM mat i j)
			 (t1,terms') = bestParse' (leftContext cd)
			                          i
					          (divisionPoint cd)
					          terms
			 (t2,terms'') = bestParse' (rightContext cd)
					           (divisionPoint cd)
						   j
					           terms'
                     in (Branch c1 [t1, t2], terms'')
  in fst (bestParse' (bestRootContext cell0) 0 (sizeTM mat) terminals)


main = do
   input <- getContents
   let world = buildMatrix (parse (lex input))
   putStr (unLex (unParse (bestParse (world))))
