{-# OPTIONS -fglasgow-exts #-}

module Tag where
import Char(isDigit, digitToInt)

import PrelGHC

type Size = Int#

data Color = Cr | Cg | Cb | Cc | Cm | Cy | Ck | Cw
    deriving (Eq, Ord, Show, Read, Enum, Bounded)

data Tag = TB | TEM | TI | TPL | TS | TTT | TU | S !Size | C !Color

allTags = [TB, TEM, TI, TPL, TS, TTT, TU,
	   S 0#, S 1#, S 2#, S 3#, S 4#, S 5#, S 6#, S 7#, S 8#, S 9#,
	   C Cr, C Cg, C Cb, C Cc, C Cm, C Cy, C Ck, C Cw]

instance Eq Tag where
  TB  == TB  = True
  TEM == TEM = True
  TI  == TI  = True
  TPL == TPL = True
  TS  == TS  = True
  TTT == TTT = True
  TU  == TU  = True
  S x == S y = x ==# y
  C x == C y = x == y
  _   == _   = False

instance Show Tag where
  showsPrec _ TB  = showChar 'B'
  showsPrec _ TEM = showChar 'E' . showChar 'M'
  showsPrec _ TI  = showChar 'I'
  showsPrec _ TPL = showChar 'P' . showChar 'L'
  showsPrec _ TS  = showChar 'S'
  showsPrec _ TTT = showChar 'T' . showChar 'T'
  showsPrec _ TU  = showChar 'U'
  showsPrec _ (S 0#) = showChar '0'
  showsPrec _ (S 1#) = showChar '1'
  showsPrec _ (S 2#) = showChar '2'
  showsPrec _ (S 3#) = showChar '3'
  showsPrec _ (S 4#) = showChar '4'
  showsPrec _ (S 5#) = showChar '5'
  showsPrec _ (S 6#) = showChar '6'
  showsPrec _ (S 7#) = showChar '7'
  showsPrec _ (S 8#) = showChar '8'
  showsPrec _ (S 9#) = showChar '9'
  showsPrec _ (C Cr) = showChar 'r'
  showsPrec _ (C Cg) = showChar 'g'
  showsPrec _ (C Cb) = showChar 'b'
  showsPrec _ (C Cc) = showChar 'c'
  showsPrec _ (C Cm) = showChar 'm'
  showsPrec _ (C Cy) = showChar 'y'
  showsPrec _ (C Ck) = showChar 'k'
  showsPrec _ (C Cw) = showChar 'w'

instance Read Tag where
  readsPrec _ ('B':s) = [(TB, s)]
  readsPrec _ ('E':'M':s) = [(TEM, s)]
  readsPrec _ ('I':s) = [(TI, s)]
  readsPrec _ ('P':'L':s) = [(TPL, s)]
  readsPrec _ ('S':s) = [(TS, s)]
  readsPrec _ ('T':'T':s) = [(TTT, s)]
  readsPrec _ ('U':s) = [(TU, s)]
  readsPrec _ ('r':s) = [(C Cr, s)]
  readsPrec _ ('g':s) = [(C Cg, s)]
  readsPrec _ ('b':s) = [(C Cb, s)]
  readsPrec _ ('c':s) = [(C Cc, s)]
  readsPrec _ ('m':s) = [(C Cm, s)]
  readsPrec _ ('y':s) = [(C Cy, s)]
  readsPrec _ ('k':s) = [(C Ck, s)]
  readsPrec _ ('w':s) = [(C Cw, s)]
  readsPrec _ ('0':s) = [(S 0#, s)]
  readsPrec _ ('1':s) = [(S 1#, s)]
  readsPrec _ ('2':s) = [(S 2#, s)]
  readsPrec _ ('3':s) = [(S 3#, s)]
  readsPrec _ ('4':s) = [(S 4#, s)]
  readsPrec _ ('5':s) = [(S 5#, s)]
  readsPrec _ ('6':s) = [(S 6#, s)]
  readsPrec _ ('7':s) = [(S 7#, s)]
  readsPrec _ ('8':s) = [(S 8#, s)]
  readsPrec _ ('9':s) = [(S 9#, s)]
  readsPrec _ _ = []

