{-# OPTIONS -fglasgow-exts #-}

module TriMatrix(
  emptyTM,
  sizeTM,
  lookupTM,
  insertTM,
  TriMatrix
) where

import Cell -- So we can specialize

data TMatrix a = Empty |
		 Tri1 !a |
		 Tris (TMatrix a) (TMatrix a) (TMatrix a) (TMatrix a)
   deriving Show

data TriMatrix a = TM {mat :: !(TMatrix a), size :: !Int, empty :: !a}
   deriving Show

emptyTM :: Int -> a -> TriMatrix a
emptyTM s zero = TM Empty s zero

sizeTM :: TriMatrix a -> Int
sizeTM = size

indexCalc :: Int -> Int -> Int -> (# Int, Int, Int, Int #)
indexCalc s i j =
   case i of { i ->
   case s `div` 2 of { s' ->
   case s - s' of { s'' ->
   if j <= s' then (# 1, s', i, j #)
   else if i >= s'' then (# 2, s', i-s'', j-s'' #)
   else if j-i <= s' then (# 4, s''-1, j-s'-1, i #)
   else (# 3, s'', i, j-s' #)
   }}}

goodTriple :: Int -> Int -> Int -> Bool
goodTriple s i j = i >= 0 && j > i && j <= s

{-# SPECIALIZE lookupTM :: TriMatrix Cell -> Int -> Int -> Cell #-}
lookupTM :: TriMatrix a -> Int -> Int -> a
lookupTM (TM m s zero) i j = {-# SCC "lookupTM" #-} lookup'' m s i j
  where
    lookup' Empty _ _ _ = zero
    lookup' (Tri1 e)     1 0 1 = e
    lookup' (Tris t1 t2 t3 t4) s i j =
      case (indexCalc s i j) of
         (# c, s', i', j' #) -> case c of
             1 -> lookup'' t1 s' i' j'
             2 -> lookup'' t2 s' i' j'
             3 -> lookup'' t3 s' i' j'
             4 -> lookup'' t4 s' i' j'
    lookup'' m s i j = seq s $ seq i $ seq j $ lookup' m s i j

{-# SPECIALIZE insertTM :: TriMatrix Cell -> Int -> Int -> Cell -> TriMatrix Cell #-}
insertTM :: TriMatrix a -> Int -> Int -> a -> TriMatrix a
insertTM (TM m s zero) i j e = {-# SCC "insertTM" #-} TM (insert'' m s i j e) s zero
  where
    insert' Empty 1 0 1 e = Tri1 e
    insert' Empty s i j e =
      case (indexCalc s i j) of
         (# c, s', i', j' #) -> case c of
              1 -> Tris (insert'' Empty s' i' j' e) Empty Empty Empty
              2 -> Tris Empty (insert'' Empty s' i' j' e) Empty Empty
              3 -> Tris Empty Empty (insert'' Empty s' i' j' e) Empty
              4 -> Tris Empty Empty Empty (insert'' Empty s' i' j' e)
    insert' (Tri1 _)        1 0 1 e = Tri1 e
    insert' (Tris t1 t2 t3 t4) s i j e =
      case (indexCalc s i j) of
         (# c, s', i', j' #) -> case c of
            1 -> Tris (insert'' t1 s' i' j' e) t2 t3 t4
            2 -> Tris t1 (insert'' t2 s' i' j' e) t3 t4
            3 -> Tris t1 t2 (insert'' t3 s' i' j' e) t4
            4 -> Tris t1 t2 t3 (insert'' t4 s' i' j' e)
    insert'' m s i j e = seq s $ seq i $ seq j $ insert' m s i j e

{-
updateTM :: TriMatrix a -> Int -> Int -> (a -> a) -> TriMatrix a
updateTM (TM m s) i j f = write' m s i f

update' (Tri1 e) 1 0 1 f = Tri1 (f e)
update' (Tris t1 t2 t3 t4) s i j f =
  case (indexCalc s i j) of
    (1, s', i', j') -> Tris (update' t1 s' i j f) t2 t3 t4
    (2, s', i', j') -> Tris t1 (update' t2 s' i j f) t3 t4
    (3, s', i', j') -> Tris t1 t2 (update' t3 s' i j f) t4
    (4, s', i', j') -> Tris t1 t2 t3 (update' t4 s' i j f)
-}

createTest :: Int -> TriMatrix (Int, Int)
createTest s =
  let l = [(i,j) | i <- [0..s-1], j <- [i+1..s]]
      doInsert (i, j) m = insertTM m i j (i,j)
  in foldr doInsert (emptyTM s (0,0)) l

